@@ -7,9 +7,21 @@ from itertools import combinations_with_replacement, repeat, chain
 from . import raw
 
 
-__version__ = '1.6.0'
+__version__ = '1.7.0'
 
 
+def _ensure_listlike(x, dups=1):
+    """Wraps x in a list if it's not list-like."""
+    try:
+        while len(x) < dups:
+            x = list(x)
+            x.append(x[-1])
+        return x
+    except TypeError:
+        return [x] * dups
+    except IndexError:
+        return []
+
 def _check_direction(direction, pardim):
     if direction in {0, 'u', 'U'} and 0 < pardim:
         return 0
@@ -35,16 +47,77 @@ def _check_edge(edge):
 
 
 def _constructor(stream):
-    peek = stream.peek(20)
-    if not peek:
-        raise raw.EOFError('')
+    if hasattr(stream, 'readline'):
+        peek = stream.readline()
+        if not peek:
+            raise raw.EOFError('')
+    else:
+        peek = stream
     if isinstance(peek, bytes):
         peek = peek.decode('utf-8')
     if peek.startswith('# LRSPLINE SURFACE'):
         return LRSplineSurface
     if peek.startswith('# LRSPLINE VOLUME'):
         return LRSplineVolume
-    raise ValueError("Unknown LRSpline object type: '{}'".format(peek[:20]))
+    raise ValueError("Unknown LRSpline object type: '{}'".format(peek))
+
+
+def _derivative_index(d):
+    """Calculate the derivative index of 'd' (a 2-tuple or 3-tuple) using
+    LRSplines' derivative numbering scheme.  Return nderivs and index.
+    """
+    nderivs = sum(d)
+    if len(d) == 2:
+        index = sum(dd + 1 for dd in range(nderivs)) + d[1]
+        return nderivs, index
+    index = nderivs * (nderivs + 1) * (nderivs + 2) // 6
+    tgt = tuple(chain.from_iterable(repeat(i,r) for i,r in enumerate(d)))
+    index += next(i for i,t in enumerate(combinations_with_replacement(range(len(d)), nderivs)) if t == tgt)
+    return nderivs, index
+
+
+def _derivative_helper(pts, derivs, func):
+    """Helper for calculating derivatives.
+
+    Pts must be a tuple of points, or a tuple of point arrays.
+
+    Derivs must be a tuple indicating which derivative to take, or a
+    list of such tuples.
+
+    Func must be a callable accepting two or three floats (the point
+    to evaluate) and an 'nderivs' argument, returning a 1-dimensional
+    array.
+
+    - If one point, one deriv: return a 1-dimensional array (coord)
+    - If multiple points, one deriv: return a 2-dimensional array (point x coord)
+    - If one point, multiple derivs: return a 2-dimensional array (deriv x coord)
+    - If multiple points, multiple derivs: return 3-dimensional (deriv x point x coord)
+    """
+
+    singlept = not isinstance(pts[0], np.ndarray)
+    singlederiv = isinstance(derivs[0], int)
+
+    if singlederiv:
+        nderiv, index = _derivative_index(derivs)
+    else:
+        nderiv, indexes = 0, []
+        for deriv in derivs:
+            n, i = _derivative_index(deriv)
+            nderiv = max(n, nderiv)
+            indexes.append(i)
+
+    if singlept:
+        data = func(*pts, nderiv)
+        if singlederiv:
+            return data[index]
+        else:
+            return np.array([data[i] for i in indexes])
+    else:
+        data = [func(*pt, nderiv) for pt in zip(*pts)]
+        if singlederiv:
+            return np.array([d[index] for d in data])
+        else:
+            return np.array([[d[i] for d in data] for i in indexes])
 
 
 class SimpleWrapper:
@@ -75,18 +148,25 @@ class BasisFunction(SimpleWrapper):
         retval = np.array([self.w.evaluate(up, vp, True, True) for up, vp in zip(u.flat, v.flat)])
         return retval.reshape(u.shape)
 
-    def derivative(self, u, v, d=(1,1)):
-        nderivs = sum(d)
-        index = nderivs * (nderivs + 1) // 2 + d[1]
-        retval = np.array([
-            self.w.evaluate(up, vp, nderivs, True, True)[index]
-            for up, vp in zip(u.flat, v.flat)
-        ])
-        return retval.reshape(u.shape)
+    def derivative(self, *pts, d=(1,1)):
+        if self.nvariate == 2:
+            wrapper = lambda u,v,n: self.w.evaluate(u, v, n, True, True)
+        else:
+            wrapper = lambda u,v,w,n: self.w.evaluate(u, v, w, n, True, True, True )
+        return _derivative_helper(pts, d, wrapper)
+
+    def support(self):
+        for w in self.w.supportIter():
+            yield Element(self.lr, w)
 
     def __getitem__(self, idx):
         return self.w.getknots(idx)
 
+    def __eq__(self, other):
+        if not isinstance(other, BasisFunction):
+            return False
+        return self.id == other.id
+
     __call__ = evaluate
 
     def refine(self):
@@ -131,6 +211,11 @@ class Element(SimpleWrapper):
         self.lr.w.refineElement(self.id)
         self.lr.w.generateIDs()
 
+    def __eq__(self, other):
+        if not isinstance(other, Element):
+            return False
+        return self.id == other.id
+
 
 class MeshInterface(SimpleWrapper):
 
@@ -282,14 +367,12 @@ class LRSplineObject:
 
     @staticmethod
     def read_many(stream):
-        objects = []
-        while True:
-            try:
-                cls = _constructor(stream)
-                objects.append(cls(stream))
-            except raw.EOFError:
-                break
-        return objects
+        contents = stream.read()
+        splitter = b'# LRSPLINE' if isinstance(contents, bytes) else '# LRSPLINE'
+        contents = contents.split(splitter)[1:]
+        contents = [splitter + c for c in contents]
+        patches = [_constructor(c)(c) for c in contents]
+        return patches
 
     def __len__(self):
         return len(self.basis)
@@ -349,11 +432,17 @@ class LRSplineObject:
             return tuple(self.w.order(d) for d in range(self.pardim))
         return self.w.order(_check_direction(direction, self.pardim))
 
-    def knots(self, direction=None):
+    def knots(self, direction=None, with_multiplicities=False):
+        if with_multiplicities:
+            knots = self.w.getGlobalKnotVector()
+        else:
+            knots = self.w.getGlobalUniqueKnotVector()
+
         if direction is None:
-            return self.w.getGlobalUniqueKnotVector()
-        direction = _check_direction(direction, self.pardim)
-        return self.w.getGlobalUniqueKnotVector()[direction]
+            return knots
+        else:
+            direction = _check_direction(direction, self.pardim)
+            return knots[direction]
 
     def refine(self, objects, beta=None):
         if not objects:
@@ -406,7 +495,7 @@ class LRSplineSurface(LRSplineObject):
         if len(args) == 0:
             w = raw.LRSurface()
         elif isinstance(args[0], raw.LRSurface):
-            w = arg
+            w = args[0]
         elif isinstance(args[0], (io.IOBase,str,bytes)):
             w = raw.LRSurface()
             w.read(args[0])
@@ -417,7 +506,8 @@ class LRSplineSurface(LRSplineObject):
         elif len(args) == 6: # specify (n1,n2), (p1,p2) and (knot1,knot2)
             w = raw.LRSurface(args[0], args[1], args[2], args[3], args[4], args[5])
         elif len(args) == 7: # specify controlpoints in addition
-            w = raw.LRSurface(args[0], args[1], args[2], args[3], args[4], args[5], args[6])
+            cp = np.array(args[-1])
+            w = raw.LRSurface(args[0], args[1], args[2], args[3], args[4], args[5], cp.flat, len(cp[0]))
         else:
             w = raw.LRSurface()
         super().__init__(w)
@@ -437,6 +527,16 @@ class LRSplineSurface(LRSplineObject):
     def write_postscript(self, stream, **kwargs):
         return self.w.writePostscriptElements(stream, **kwargs)
 
+    def insert_knot(self, new_knots, direction, multiplicity=1):
+        new_knots = _ensure_listlike(new_knots)
+        direction = _check_direction(direction, self.pardim)
+        for k in new_knots:
+            if direction == 0:
+                self.w.insert_const_u_edge(k, self.start(1), self.end(1), multiplicity)
+            else:
+                self.w.insert_const_v_edge(k, self.start(0), self.end(0), multiplicity)
+        self.w.generateIDs()
+
     def insert(self, *args, direction=None, value=None, start=None, end=None, multiplicity=None):
         if len(args) > 1:
             raise TypeError('Expected at most one positional argument')
@@ -468,23 +568,19 @@ class LRSplineSurface(LRSplineObject):
     def evaluate(self, u, v, iel=-1):
         if isinstance(u, np.ndarray) and isinstance(v, np.ndarray):
             retval = np.array([self.w.point(up, vp, iEl=iel) for up, vp in zip(u.flat, v.flat)])
-            return retval.reshape(u.shape)
+            return retval.reshape(u.shape + (-1,))
         return self.w.point(u, v, iEl=iel)
 
     def derivative(self, u, v, d=(1,1), iel=-1):
-        nderivs = sum(d)
-        index = sum(dd + 1 for dd in range(nderivs)) + d[1]
-        if isinstance(u, np.ndarray) and isinstance(v, np.ndarray):
-            retval = []
-            for up, vp in zip(u.flat, v.flat):
-                r = self.w.point(up, vp, nderivs, iEl=iel)
-                retval.append(r[index])
-            return np.array(retval).reshape(u.shape)
-        return self.w.point(u, v, nderivs, iEl=iel)[index]
+        wrapper = lambda u,v,n: self.w.point(u, v, n, iEl=iel)
+        return _derivative_helper((u, v), d, wrapper)
 
     def bezier_extraction(self, iEl):
         return self.w.getBezierExtraction(iEl)
 
+    def element_at(self, u, v):
+        return self.elements[self.w.getElementContaining(u,v)]
+
     __call__ = evaluate
 
 
@@ -494,7 +590,7 @@ class LRSplineVolume(LRSplineObject):
         if len(args) == 0:
             w = raw.LRVolume()
         elif isinstance(args[0], raw.LRVolume):
-            w = arg
+            w = args[0]
         elif isinstance(args[0], (io.IOBase,str,bytes)):
             w = raw.LRVolume()
             w.read(args[0])
@@ -505,7 +601,8 @@ class LRSplineVolume(LRSplineObject):
         elif len(args) == 9: # specify n,p and knotvector for 3 directions
             w = raw.LRVolume(*args)
         elif len(args) == 10: # specify all above in addition to controlpoints
-            w = raw.LRVolume(*args)
+            cp = np.array(args[-1])
+            w = raw.LRVolume(*args[:-1], cp.flat, len(cp[0]))
         else:
             w = raw.LRVolume()
         super().__init__(w)
@@ -526,29 +623,34 @@ class LRSplineVolume(LRSplineObject):
     def clone(self):
         return LRSplineVolume(self.w.copy())
 
+    def insert_knot(self, new_knots, direction, multiplicity=1):
+        new_knots = _ensure_listlike(new_knots)
+        direction = _check_direction(direction, self.pardim)
+        for k in new_knots:
+            start = list(self.start())
+            end = list(self.end())
+            start[direction] = end[direction] = k
+            mr = raw.MeshRectangle(start[0], start[1], start[2], end[0], end[1], end[2], multiplicity)
+            self.w.insert_line(mr)
+        self.w.generateIDs()
+
     def insert(self, mr):
-        self.w.insert(mr)
+        self.w.insert_line(mr.w.copy())
 
     def evaluate(self, u, v, w, iel=-1):
-        if isinstance(u, np.ndarray) and isinstance(u, np.ndarray) and isinstance(u, np.ndarray):
+        if isinstance(u, np.ndarray) and isinstance(v, np.ndarray) and isinstance(w, np.ndarray):
             retval = np.array([self.w.point(up, vp, wp, iEl=iel) for up, vp, wp in zip(u.flat, v.flat, w.flat)])
-            return retval.reshape(u.shape)
+            return retval.reshape(u.shape + (-1,))
         return self.w.point(u, v, w, iEl=iel)
 
     def bezier_extraction(self, iEl):
         return self.w.getBezierExtraction(iEl)
 
+    def element_at(self, u, v, w):
+        return self.elements[self.w.getElementContaining(u,v,w)]
+
     def derivative(self, u, v, w, d=(1,1,1), iel=-1):
-        nderivs = sum(d)
-        index = nderivs * (nderivs + 1) * (nderivs + 2) // 6
-        tgt = tuple(chain.from_iterable(repeat(i,r) for i,r in enumerate(d)))
-        index += next(i for i,t in enumerate(combinations_with_replacement(range(len(d)), nderivs)) if t == tgt)
-        if isinstance(u, np.ndarray) and isinstance(u, np.ndarray) and isinstance(u, np.ndarray):
-            retval = []
-            for up, vp, wp in zip(u.flat, v.flat, w.flat):
-                r = self.w.point(up, vp, wp, nderivs, iEl=iel)
-                retval.append(r[index])
-            return np.array(retval).reshape(u.shape)
-        return self.w.point(u, v, w, nderivs, iEl=iel)[index]
+        wrapper = lambda u,v,w,n: self.w.point(u, v, w, n, iEl=iel)
+        return _derivative_helper((u, v, w), d, wrapper)
 
     __call__ = evaluate
